/*
 *
 * Copyright (c) 2001
 *	Politecnico di Torino.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Author:	Marco Mellia, Andrea Carpani, Luca Muscariello, Dario Rossi
 * 		Telecomunication Networks Group
 * 		Politecnico di Torino
 * 		Torino, Italy
 *              http://http://www.tlc-networks.polito.it/index.html
 *		mellia@mail.tlc.polito.it
 *
 * Tstat is deeply based on TCPTRACE. The authors would like to thank
 * Shawn Ostermann for the development of TCPTRACE.
 *
*/

static char const copyright[] =
  "@(#)Copyright (c) 2001 -- Telecomunication Network Group \
     -- Politecnico di Torino.  All rights reserved.\
     Tstat is based on TCPTRACE,\
    @(#)Copyright (c) 1999 -- Shawn Ostermann -- Ohio University.\n";

#include "tstat.h"
#include "file_formats.h"
#include "version.h"
#include <time.h>

/* version information */
char *tstat_version = VERSION;

/* seem to be missing from pthread.h */
int pthread_mutexattr_settype (pthread_mutexattr_t * attr, int kind);

/* local routines */
void QuitSig (int);
void Usr1Sig (int);
static void Formats (void);
static void Help ();
static void CheckArguments (int *pargc, char *argv[]);
static void ParseArgs (char *argsource, int *pargc, char *argv[]);
static void ProcessFile (char *filename);
static void Usage (void);
static void BadArg (char *argsource, char *format, ...);
static void Version (void);
static int LoadInternalNets (char *file);

static Bool internal_ip (struct in_addr adx);
static Bool internal_ip_string (char *adx);

static int ip_header_stat (int phystype, struct ip *pip, u_long * fpnum,
			   u_long * pcount, int file_count, char *filename,
			   long int location, int tlen);
struct tcphdr *tcp_header_stat (struct tcphdr *ptcp, struct ip *pip,
				void *plast);

void stat_dumping_old_style ();

/* thread mutex and conditional variables  */

Bool threaded = FALSE;
pthread_mutex_t ttp_lock_mutex;
pthread_mutex_t utp_lock_mutex;
pthread_mutex_t flow_close_started_mutex;
pthread_mutex_t flow_close_cond_mutex;
pthread_cond_t flow_close_cond;
pthread_mutex_t stat_dump_cond_mutex;
pthread_mutex_t stat_dump_mutex;
pthread_cond_t stat_dump_cond;


Bool filter_specified = FALSE;
char *filter_filename;
char *dev;

static Bool basedirspecified = FALSE;
static char *basenamedir;
Bool internal_src = FALSE;
Bool internal_dst = FALSE;

struct in_addr internal_net_list[MAX_INTERNAL_HOSTS];
struct in_addr internal_net_mask2[MAX_INTERNAL_HOSTS];
int internal_net_mask[MAX_INTERNAL_HOSTS];
char *internal_net_file;
int tot_internal_nets;

/* option flags and default values */
Bool live_flag = FALSE;

Bool printticks = FALSE;
Bool dump_all_histo_definition = FALSE;

Bool warn_IN_OUT = TRUE;
Bool warn_ooo = FALSE;
Bool warn_MAX_ = TRUE;
Bool warn_printtrunc = FALSE;
Bool warn_printbadmbz = FALSE;
Bool warn_printbadcsum = FALSE;
Bool warn_printbad_syn_fin_seq = FALSE;
Bool do_udp = TRUE;
int debug = 0;
char **filenames = NULL;
int two_files = 1;		/* used if you have two traces that store packet in
				   two different direction... works only with DAG for now.
				 */

u_long pnum = 0;
u_long ctrunc = 0;
u_long bad_ip_checksums = 0;
u_long bad_tcp_checksums = 0;
u_long bad_udp_checksums = 0;

/* globals */
int *coredump;
struct timeval current_time;

struct adx *adx_index_first[MAX_ADX_SLOTS];
struct adx *adx_index_second[MAX_ADX_SLOTS];
struct adx **adx_index_current;

unsigned long int fcount = 0;	/* total flow number */
unsigned long int f_TCP_count = 0;	/* total TCP  flow number */
unsigned long int f_UDP_count = 0;	/* total UDP flow number */
unsigned long int f_RTP_count = 0;	/* total RTP flow number */
unsigned long int f_RTCP_count = 0;	/* total RTP flow number */

#ifdef HAVE_RRDTOOL
/*-----------------------------------------------------------*/
/* RRDtools 				                     */
/*-----------------------------------------------------------*/
Bool rrdset_path = FALSE;	/* database path flag */
Bool rrdset_conf = FALSE;	/* configuration file flag */
Bool pure_rrd_engine = FALSE;	/* -S */
/*-----------------------------------------------------------*/
#endif

/* locally global variables */
static u_long filesize = 0;
static int num_files = 0;
static u_int numfiles;
static char *cur_filename;
static char *progname;

//XXX
typedef enum
{ ETH, DAG } Live_Cap_Type;
int livecap_type;		/* indicate the type of live capture */
#ifdef GROK_ERF_LIVE
char *dag_dev_list;		/* list of DAG cards device names */
#define DAG_NAME_BUFSIZE 25
#endif

/* for elapsed processing time */
struct timeval wallclock_start;
struct timeval wallclock_finished;
struct timeval wallclock_temp;

/* first and last packet timestamp */
timeval first_packet = { 0, 0 };
timeval last_packet = { 0, 0 };

/* .a.c. */
Bool is_stdin;
FILE *fp_logc;
FILE *fp_lognc;
FILE *fp_rtp_logc;

/* LM */
#ifdef LOG_OOO
FILE *fp_dup_ooo_log;
#endif

#ifdef INTERNAL_WIRED
Bool coming_in;
#endif

extern long int tcp_packet_count;
extern long not_id_p;
extern int search_count;
extern long int tot_adx_hash_count, tot_adx_list_count, adx_search_hash_count,
  adx_search_list_count;

static void
Help (void)
{
  fprintf (stderr, "Usage: tstat  [-hHtuvwp] [-d[-d]] [-Nfile] [-sdir]\n");
  fprintf (stderr,
#ifdef HAVE_RRDTOOL
	   "              [-rRRD_out_dir] [-Rrrd_conf] [-S]\n"
#endif
#ifdef GROK_LIVE_TCPDUMP
	   "              [-l] [-iinterface]\n"
#endif
#ifdef GROK_ERF_LIVE
	   "              [-dag device_name device_name ...]\n"
#endif
	   "              [-ffilterfile] <file1 file2>\n\n");
  fprintf (stderr, "\t-h: print this help and exit\n");
  fprintf (stderr, "\t-H: print internal histograms names and definitions\n");
  fprintf (stderr, "\t-t: print ticks showing the trace analysis progress\n");
  fprintf (stderr, "\t-u: do not trace UDP packets\n");
  fprintf (stderr, "\t-v: print version and exit\n");
  fprintf (stderr, "\t-w: print [lots] of warning\n");
  fprintf (stderr,
	   "\t-p: enable multi-threaded engine (useful for live capture)\n\n");
  fprintf (stderr,
	   "\t-d: increase debug level (repeat to increase debug level)\n");
  fprintf (stderr, "\t-Nfile: specify the file name which contains the\n");
  fprintf (stderr, "\t        description of the internal networks.\n");
  fprintf (stderr,
	   "\t        This file must contain the subnets that will be\n");
  fprintf (stderr,
	   "\t        considered as 'internal' during the analysis\n");
  fprintf (stderr,
	   "\t        Each subnet must be specified using network IP address\n");
  fprintf (stderr,
	   "\t        on the first line and NETMASK on the next line:\n");
  fprintf (stderr, "\t        130.192.0.0\n");
  fprintf (stderr, "\t        255.255.0.0\n");
  fprintf (stderr, "\t        193.204.134.0\n");
  fprintf (stderr, "\t        255.255.255.0\n\n");

  fprintf (stderr, "\t-sdir: puts the trace analysis results into directory\n");
  fprintf (stderr, "\t       tree dir (otherwise will be <file>.out)\n");

#ifdef HAVE_RRDTOOL
  fprintf (stderr,
	   "\t-S: pure RRD-engine: do not create histograms and log_* files \n");
/*----------------------------------------------------------- 
   RRDtools 				                     
   these flags test for both the -r and -R options to be 
   specified when using RR database integration */
  fprintf (stderr,
	   "\t-Rconf: specify the configuration file for integration with"
	   "\n\t      RRDtool. See README.RRDtool for further information\n");
  fprintf (stderr,
	   "\t-rpath: path to use to create/update the RRDtool database: this"
	   "\n\t      should better be outside the directory tree"
	   "\n\t      and should be accessible from the Web server\n"
	   "\n\t      running the cgi script interface\n");
/*-----------------------------------------------------------*/
#endif
#ifdef GROK_LIVE_TCPDUMP
  fprintf (stderr, "\t-l: enable live capture using libpcap\n");
  fprintf (stderr,
	   "\t-iinterface: specifies the interface to be used to capture traffic\n");
#endif /* GROK_LIVE_TCPDUMP */

#ifdef GROK_ERF_LIVE
  fprintf (stderr,
	   "\t-dag: enable live capture using Endace DAG cards. The default device for capture is\n"
	   "\t      /dev/dag0 and the list of device_name can contain at most four name\n");
#endif /* GROK_ERF_LIVE */

  fprintf (stderr,
	   "\t-ffiterfile: specifies the libpcap filter file. Syntax as in tcpdump\n\n");
  fprintf (stderr, "\tfile: trace file to be analyzed\n");
  fprintf (stderr, "\t      Use 'stdin' to read from standard input.\n");

  fprintf (stderr, "\n");
  Formats ();
  Version ();
}



static void
BadArg (char *argsource, char *format, ...)
{
  va_list ap;

  Help ();

  fprintf (stderr, "\nArgument error");
  if (argsource)
    fprintf (stderr, " (from %s)", argsource);
  fprintf (stderr, ": ");

  va_start (ap, format);
  vfprintf (stderr, format, ap);
  va_end (ap);
  fprintf (stderr, "\n");
  exit (EXIT_FAILURE);

}



static void
Usage (void)
{
  Help ();
  exit (0);
}



static void
Version (void)
{
  fprintf (stderr, "Version: %s\n", tstat_version);
  fprintf (stderr, "  Compiled by '%s' at '%s' on machine '%s'\n",
	   built_bywhom, built_when, built_where);
}


static void
Formats (void)
{
  int i;

  fprintf (stderr, "Supported Input File Formats:\n");
  for (i = 0; i < (int) NUM_FILE_FORMATS; ++i)
    fprintf (stderr, "\t%-15s  %s\n",
	     file_formats[i].format_name, file_formats[i].format_descr);
}


int
main (int argc, char *argv[])
{
  int i, j;
  double etime;
  pthread_t thread_done_periodic;
  pthread_t thread_all_dumping;
  pthread_mutexattr_t attr;

  if (argc == 1)
    {
      Help ();
      exit (1);
    }

  /* initialize internals */
  trace_init ();

  /* parse the flags */
  CheckArguments (&argc, argv);

  /* optional UDP */
  if (do_udp)
    udptrace_init ();

  /* get starting wallclock time */
  gettimeofday (&wallclock_start, NULL);

/* allocate all histo structs */
  create_all_histo ();

  if (dump_all_histo_definition == TRUE)
    {
      print_all_histo_definition ();
      printf ("\n");
      exit (0);
    }

  if (live_flag == FALSE)
    {				/* se  una cattura live non ci sono file quindi questo no serve */
      num_files = argc;
      printf ("%d arg%s remaining, starting with '%s'\n",
	      num_files, num_files > 1 ? "s" : "", filenames[0]);
    }

  /* knock, knock... */
  printf ("%s\n\n", VERSION);


#ifdef HAVE_RRDTOOL
  /*-----------------------------------------------------------*/
  /* RRDtools                                                   */
  /*   now that all the histo have been creaed, we may          */
  /*   parse rrdtool configuration file                         */
  if (rrdset_path && rrdset_conf)
    rrdtool_init ();
  /*-----------------------------------------------------------*/
#endif

/* inititializing adx_index_current */
  adx_index_current = adx_index_first;

/* thread creation and management  */

  if (threaded)
    {
      /* Initialize mutex and condition variable objects */
      pthread_mutexattr_init (&attr);
      pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_ERRORCHECK_NP);
      pthread_mutex_init (&ttp_lock_mutex, &attr);


      /* Inizializzazione di attr2 e del mutex relativo ai flussi udp */
      pthread_mutex_init (&utp_lock_mutex, NULL);


      pthread_mutex_init (&flow_close_cond_mutex, NULL);
      pthread_mutex_init (&flow_close_started_mutex, NULL);
      pthread_cond_init (&flow_close_cond, NULL);

      pthread_mutex_init (&stat_dump_mutex, NULL);
      pthread_mutex_init (&stat_dump_cond_mutex, NULL);
      pthread_cond_init (&stat_dump_cond, NULL);
      /* flow_closing thread */
      pthread_create (&thread_done_periodic, NULL, time_out_flow_closing,
		      NULL);
      /* stat_dump thread */
      pthread_create (&thread_all_dumping, NULL, stats_dumping, NULL);
      /* allow the thread to start... */
      usleep (200);
    }

  /* read each file in turn */
  if (live_flag == FALSE)
    {
      numfiles = argc;
      for (i = 0; i < argc; i++)
	{
	  for (j = 0; j < two_files; j++)
	    {
	      if ((debug > 0) || (numfiles > 1))
		{
		  if (argc > 1)
		    printf ("Running file '%s' (%d of %d)\n",
			    filenames[i + j], i + j + 1, numfiles);
		  else
		    printf ("Running file '%s'\n", filenames[i]);
		}
	    }

	  /* do the real work */
	  ProcessFile (filenames[i]);
	  i += j - 1;
	}
    }
  else
    {
      ProcessFile ("stdin");	/* nel caso sia cattura live non devo passargli nessun file */
    }


  if (threaded)
    {
      /* Closing flow_closing thread */
      pthread_cancel (thread_done_periodic);
      pthread_cancel (thread_all_dumping);
    }

  /* clean up output */
  if (printticks)
    printf ("\n");

  /* get ending wallclock time */
  gettimeofday (&wallclock_finished, NULL);

  /* general output */
  fprintf (stdout, "%lu packets seen, %lu TCP packets traced",
	   pnum, tcp_trace_count_outgoing + tcp_trace_count_incoming
	   + tcp_trace_count_local);
  if (do_udp)
    fprintf (stdout, ", %lu UDP packets traced", udp_trace_count);
  fprintf (stdout, "\n");

  /* actual tracefile times */
  etime = elapsed (first_packet, last_packet);
  fprintf (stdout, "trace %s elapsed time: %s\n",
	   (num_files == 1) ? "file" : "files", elapsed2str (etime));
  if (debug > 0)
    {
      fprintf (stdout, "\tfirst packet:  %s\n", ts2ascii (&first_packet));
      fprintf (stdout, "\tlast packet:   %s\n", ts2ascii (&last_packet));
    }
  exit (EXIT_SUCCESS);
}


// MGM
/* Create subdirs into which out files will be put */
void
create_new_outfiles (char *filename)
{
  char tmpstr[200];
  struct stat fbuf;
  char date[50];
  char logfile[200] = "";

#ifdef HAVE_RRDTOOL
  if (pure_rrd_engine)
    return;
#endif

  if (!basedirspecified)
    {
      /* get the basename from the tracefile */
      if (is_stdin)
	{
	  basenamedir = strdup ("stdin");
	}
      else
	{
	  basenamedir = get_basename (filename);
	}
    }
  if (stat (basenamedir, &fbuf) != 0)
    {
      mkdir (basenamedir, 0775);
    }
  basename = (char *) malloc (100 * sizeof (char));
  strftime (date, 49, "%H_%M_%d_%b_%Y", localtime (&current_time.tv_sec));

  sprintf (basename, "%s/%s.out", basenamedir, date);

  if (stat (basename, &fbuf) != -1)
    {
      /* remove the previous directory */
      sprintf (tmpstr, "rm -rf %s", basename);
      system (tmpstr);
    }
  mkdir (basename, 0775);
  sprintf (curr_data_dir, "%s/LAST", basename);

  /* Open the files for complete and uncomplete connection logging */

  sprintf (logfile, "%s/%s", basename, "log_tcp_complete");
  if (fp_logc != NULL)
    fclose (fp_logc);
  fp_logc = fopen (logfile, "w");
  if (fp_logc == NULL)
    {
      printf ("Could not open file %s\n", logfile);
    }

  sprintf (logfile, "%s/%s", basename, "log_tcp_nocomplete");
  if (fp_lognc != NULL)
    fclose (fp_lognc);
  fp_lognc = fopen (logfile, "w");
  if (fp_lognc == NULL)
    {
      printf ("Could not open file %s\n", logfile);
    }

/* crea file per sessioni RTP completate */

  sprintf (logfile, "%s/%s", basename, "log_rtp_complete");
  if (fp_rtp_logc != NULL)
    fclose (fp_rtp_logc);
  fp_rtp_logc = fopen (logfile, "w");
  if (fp_rtp_logc == NULL)
    {
      printf ("Could not open file %s\n", logfile);
    }

#ifdef LOG_OOO
/* MGM start */
  /* Open the files for dup and ooo logging */

  sprintf (logfile, "%s/%s", basename, "dup_ooo");
  if (fp_dup_ooo_log != NULL)
    fclose (fp_dup_ooo_log);
  fp_dup_ooo_log = fopen (logfile, "w");
  if (fp_dup_ooo_log == NULL)
    {
      printf ("Could not open file %s\n", logfile);
    }
/* MGM stop */
#endif
}



inline static int
ip_header_stat (int phystype, struct ip *pip, u_long * fpnum, u_long * pcount,
		int file_count, char *filename, long int location, int tlen)
{

  /* another sanity check, only understand ETHERNET right now */

  if (phystype != PHYS_ETHER)
    {
      static int not_ether = 0;

      ++not_ether;
      if (not_ether < 5)
	{
	  fprintf (stderr,
		   "Skipping packet %lu, not an ethernet packet\n", pnum);
	}			/* else, just shut up */
      return 0;
    }

  /* decide wheater this is internal or external */
#ifdef INTERNAL_WIRED
  internal_src = coming_in;
  internal_dst = !coming_in;
#else
  internal_src = internal_ip (pip->ip_src);
  internal_dst = internal_ip (pip->ip_dst);
#endif

  /* .a.c. */
  if (internal_src && !internal_dst)
    {
      add_histo (ip_protocol_out, pip->ip_p);
      add_histo (ip_len_out, (float) ntohs (pip->ip_len));
      add_histo (ip_ttl_out, (float) pip->ip_ttl);
      add_histo (ip_tos_out, (float) pip->ip_tos);
    }
  else if (!internal_src && internal_dst)
    {
      add_histo (ip_protocol_in, pip->ip_p);
      add_histo (ip_len_in, (float) ntohs (pip->ip_len));
      add_histo (ip_ttl_in, (float) pip->ip_ttl);
      add_histo (ip_tos_in, (float) pip->ip_tos);
    }
  else if (internal_src && internal_dst)
    {
      add_histo (ip_protocol_loc, pip->ip_p);
      add_histo (ip_len_loc, (float) ntohs (pip->ip_len));
      add_histo (ip_ttl_loc, (float) pip->ip_ttl);
      add_histo (ip_tos_loc, (float) pip->ip_tos);
    }


  add_adx (&(pip->ip_src), SRC_ADX);
  add_adx (&(pip->ip_dst), DST_ADX);

  /* update global and per-file packet counters */
  ++pnum;			/* global */
  ++(*fpnum);			/* local to this file */
  ++(*pcount);			/* counter per chiudere i pendenti */


  /* the last_time_step is assigned only at the first packet */
  if (pnum == 1)
    last_time_step = last_cleaned = current_time;


  /* check for re-ordered packets */
  if (!ZERO_TIME (&last_packet))
    {
      if (elapsed (last_packet, current_time) < 0)
	{
	  /* out of order */
	  if ((file_count > 1) && ((*fpnum) == 1))
	    {
	      fprintf (stderr, "\
Warning, first packet in file %s comes BEFORE the last packet\n\
in the previous file.  That will likely confuse the program, please\n\
order the files in time if you have trouble\n", filename);
	    }
	  else
	    {
	      static int warned = 0;

	      if (warn_ooo)
		{
		  fprintf (stderr, "\
Warning, packet %ld in file %s comes BEFORE the previous packet\n\
That will likely confuse the program, so be careful!\n", (*fpnum), filename);
		}
	      else if (!warned)
		{
		  fprintf (stderr, "\
Packets in file %s are out of order.\n\
That will likely confuse the program, so be careful!\n", filename);
		}
	      warned = 1;
	    }

	}
    }


  /* install signal handler */
  if ((*fpnum) == 1)
    {
      signal (SIGINT, QuitSig);
      signal (SIGUSR1, Usr1Sig);
    }


  /* progress counters */
  if (printticks)
    {
      if (CompIsCompressed ())
	location += tlen;	/* just guess... */
      if ((((*fpnum) < 100) && ((*fpnum) % 10 == 0)) ||
	  (((*fpnum) < 1000) && ((*fpnum) % 100 == 0)) ||
	  (((*fpnum) < 10000) && ((*fpnum) % 1000 == 0)) ||
	  (((*fpnum) < 100000) && ((*fpnum) % 10000 == 0)) ||
	  (((*fpnum) >= 100000) && ((*fpnum) % 100000 == 0)))
	{

	  unsigned frac;

	  if (debug)
	    fprintf (stderr, "%s: ", cur_filename);
	  fprintf (stderr, "Tp= %lu Tf=%lu ", (*fpnum), fcount);
	  if (CompIsCompressed ())
	    {
	      frac = location / filesize * 100;
	      if (frac <= 100)
		fprintf (stderr, "~%u%% (compressed)", frac);
	      else
		fprintf (stderr, "~100%% + %u%% (compressed)", frac - 100);
	    }
	  else if (!is_stdin)
	    {
	      location = ftell (stdin);
	      frac = location / filesize * 100;
	      fprintf (stderr, "%u%%", frac);
	    }
	  /* print elapsed time */
	  {
	    double etime = elapsed (first_packet, last_packet);
	    fprintf (stderr, " (%s)", elapsed2str (etime));
	  }
	  /* print number of opened flow */
	  {
	    fprintf (stderr, " Nf(TCP)=%lu Nf(UDP)=%lu", tot_conn_TCP,
		     tot_conn_UDP);
	    fprintf (stderr, " Ntrash=%lu", not_id_p);
	  }

	  /* carriage return (but not newline) */
	  fprintf (stderr, "\r");
	}
      fflush (stderr);
    }

  /* keep track of global times */
  if (ZERO_TIME (&first_packet))
    first_packet = current_time;
  last_packet = current_time;

  return 1;			/*finished ok */
}

inline struct tcphdr *
tcp_header_stat (struct tcphdr *ptcp, struct ip *pip, void *plast)
{

  /* find the start of the TCP header */
  ptcp = gettcp (pip, &plast);

  /* if that failed, it's not TCP */
  if (ptcp == NULL)
    {
      udp_pair *pup;
      struct udphdr *pudp;

      /* look for a UDP header */
      pudp = getudp (pip, &plast);

      if (do_udp && (pudp != NULL))
	{
	  pup = udpdotrace (pip, pudp, plast);

	}
      else if (pudp == NULL)
	{
	}
      return 0;
    }

  /* perform TCP packet analysis */

  if ((!ACK_SET (ptcp) && SYN_SET (ptcp)))
    {
      if (internal_src && !internal_dst)
	{
	  add_histo (tcp_port_syn_src_out, (float) ntohs (ptcp->th_sport));
	  add_histo (tcp_port_syn_dst_out, (float) ntohs (ptcp->th_dport));
	}
      else if (!internal_src && internal_dst)
	{
	  add_histo (tcp_port_syn_src_in, (float) ntohs (ptcp->th_sport));
	  add_histo (tcp_port_syn_dst_in, (float) ntohs (ptcp->th_dport));
	}
      else if (internal_src && internal_dst)
	{
	  add_histo (tcp_port_syn_src_loc, (float) ntohs (ptcp->th_sport));
	  add_histo (tcp_port_syn_dst_loc, (float) ntohs (ptcp->th_dport));
	}
    }

  if (internal_src && !internal_dst)
    {
      add_histo (tcp_port_src_out, (float) ntohs (ptcp->th_sport));
      add_histo (tcp_port_dst_out, (float) ntohs (ptcp->th_dport));
    }
  else if (!internal_src && internal_dst)
    {
      add_histo (tcp_port_src_in, (float) ntohs (ptcp->th_sport));
      add_histo (tcp_port_dst_in, (float) ntohs (ptcp->th_dport));
    }
  else if (internal_src && internal_dst)
    {
      add_histo (tcp_port_src_loc, (float) ntohs (ptcp->th_sport));
      add_histo (tcp_port_dst_loc, (float) ntohs (ptcp->th_dport));
    }

  return ptcp;
}

static void
ProcessFile (char *filename)
{
  pread_f *ppread = NULL;
  int ret = 0;
  struct ip *pip;
  struct tcphdr *ptcp = NULL;
  int phystype;
  void *phys;			/* physical transport header */
  tcp_pair *ptp;
  int fix;
  int len;
  int tlen;
  void *plast;
  struct stat str_stat;
  long int location = 0;
  u_long fpnum = 0;
/* used to count the opened flows...*/
  u_long pcount = 0;
  static int file_count = 0;


  /* share the current file name */
  cur_filename = filename;

  /* open the file header */
  if (CompOpenHeader (filename) == NULL)
    {
      exit (EXIT_FAILURE);
    }

#ifdef __WIN32
  /* If the file is compressed, exit (Windows version does not support compressed dump files) */
  if (CompOpenHeader (filename) == (FILE *) - 1)
    {
      exit (-1);
    }
#else
  /* open the file header */
  if (CompOpenHeader (filename) == NULL)
    {
      exit (-1);
    }
#endif /* __WIN32 */

  /* see how big the file is */
  is_stdin = FALSE;
  filesize = 1;
  if (FileIsStdin (filename))
    {
      filesize = 1;
      is_stdin = TRUE;
    }
  else
    {
      if (stat (filename, &str_stat) != 0)
	{
	  perror ("stat");
	  exit (EXIT_FAILURE);
	}
      else
      {
        if(str_stat.st_mode == S_IFREG)
  	   filesize = str_stat.st_size;
      }
      
    }

  if (live_flag == TRUE)
    {				/*is a live capture */
      switch (livecap_type)
	{
#ifdef GROK_LIVE_TCPDUMP
	case ETH:
	  ppread = (*file_formats[ETH_LIVE].test_func) (filename);
	  if (debug > 0)
	    fprintf (stderr, "Capturing using '%s' (%s)\n",
		     file_formats[ETH_LIVE].format_name,
		     file_formats[ETH_LIVE].format_descr);
	  break;
#endif
  
#ifdef GROK_ERF_LIVE
	case DAG:
	  ppread = (*file_formats[ERF_LIVE].test_func) (dag_dev_list);
	  free (dag_dev_list);
	  if (debug > 0)
	    fprintf (stderr, "Capturing using '%s' (%s)\n",
		     file_formats[ERF_LIVE].format_name,
		     file_formats[ERF_LIVE].format_descr);
	  break;
#endif
	}
    }

  else
    {
      /* determine which input file format it is... */
      ppread = NULL;
      if (debug > 1)
	printf ("NUM_FILE_FORMATS: %d\n", (int) NUM_FILE_FORMATS);
      for (fix = 0; fix < (int) NUM_FILE_FORMATS - NUM_LIVE_FORMATS; ++fix)
	{
	  if (debug > 0)
	    fprintf (stderr, "Checking for file format '%s' (%s)\n",
		     file_formats[fix].format_name,
		     file_formats[fix].format_descr);
	  rewind (stdin);
	  ppread = (*file_formats[fix].test_func) (filename);	/* determine the
								   input file format */
	  if (ppread)
	    {
	      if (debug > 0)
		fprintf (stderr, "File format is '%s' (%s)\n",
			 file_formats[fix].format_name,
			 file_formats[fix].format_descr);
	      break;
	    }
	  else if (debug > 0)
	    {
	      fprintf (stderr, "File format is NOT '%s'\n",
		       file_formats[fix].format_name);
	    }
	}

      /* if we haven't found a reader, then we can't continue */
      if (ppread == NULL)
	{
	  int count = 0;

	  fprintf (stderr, "Unknown input file format\n");
	  Formats ();

	  /* check for ASCII, a common problem */
	  rewind (stdin);
	  while (TRUE)
	    {
	      int ch;
	      if ((ch = getchar ()) == EOF)
		break;
	      if (!isprint (ch))
		break;
	      if (++count >= 20)
		{
		  /* first 20 are all ASCII */
		  fprintf (stderr,
			   "\nThis looks like an ASCII input file to me.\n");
		  exit (EXIT_FAILURE);
		}
	    }
	  exit (EXIT_FAILURE);
	}

#ifndef __WIN32
      /* open the file for processing */
      if (CompOpenFile (filename) == NULL)
	{
	  exit (-1);
	}
#endif /* __WIN32 */

      /* how big is it.... (possibly compressed) */
      if (debug > 0)
	{
	  /* print file size */
	  printf ("Trace file size: %lu bytes\n", filesize);
	}
      location = 0;


      /* count the files */
      ++file_count;

    }			/************************end else di if (filename==NULL) per la cattura live *****************/

  // MGM
  /* read the first packet, to get the timestamp of the trace */
  while (current_time.tv_sec == 0 && current_time.tv_usec == 0)
    {
      ret = (*ppread) (&current_time, &len, &tlen, &phys, &phystype, &pip,
		       &plast);
    }
  /* and then rewind the stream... if it is not DAG or TCPDUMP format */
#ifdef GROK_DAG
  if (ppread != pread_dag)
#else
  #ifdef GROK_LIVE_TCPDUMP
  if (ppread != pread_tcpdump)
  #endif
#endif
    rewind (stdin);
  create_new_outfiles (filename);

  tot_conn_TCP = 0;
  tot_conn_UDP = 0;

  /* read each packet */
  while (TRUE)
    {
      /* read the next packet */
      ret =
	(*ppread) (&current_time, &len, &tlen, &phys, &phystype, &pip,
		   &plast);


      if (ret == 0)		/* EOF */
	break;

      /* quick sanity check, better be an IPv4/v6 packet */
      if (!PIP_ISV4 (pip) && !PIP_ISV6 (pip))
	{
	  static Bool warned = FALSE;

	  if (!warned)
	    {
	      fprintf (stderr, "Warning: saw at least one non-ip packet\n");
	      warned = TRUE;
	    }

	  if (debug > 1)
#ifdef SUPPORT_IPV6
	    fprintf (stderr,
		     "Skipping packet %lu, not an IPv4/v6 packet (version:%d)\n",
		     pnum, pip->ip_v);
#else
	    fprintf (stderr,
		     "Skipping packet %lu, not an IPv4 packet (version:%d)\n",
		     pnum, pip->ip_v);
#endif
	  continue;
	}




/* Statistics from IP HEADER */
      if (ip_header_stat
	  (phystype, pip, &fpnum, &pcount, file_count, filename, location,
	   tlen) == 0)
	continue;

      if (elapsed (last_time_step, current_time) > MAX_TIME_STEP)
	{
	  last_time_step = current_time;

	  if (threaded)
	    {
#ifdef DEBUG_THREAD
	      printf ("Signaling thread stat dump\n");
#endif
	      pthread_mutex_lock (&stat_dump_cond_mutex);
	      pthread_cond_signal (&stat_dump_cond);
	      pthread_mutex_unlock (&stat_dump_cond_mutex);
	      pthread_mutex_lock (&stat_dump_mutex);
	      pthread_mutex_unlock (&stat_dump_mutex);
	    }
	  else
	    stat_dumping_old_style ();
	}


/* Statistics from LAYER 4 (TCP/UDP) HEADER */
      if ((ptcp = tcp_header_stat (ptcp, pip, plast)) == 0)
	continue;
/* Statistics from TCP FLOWS */
      ptp = dotrace_2 (pip, ptcp, plast);

      /* if it wasn't "interesting", we return NULL here */
      if (ptp == NULL)
	continue;

      /* for efficiency, only allow a signal every 1000 packets       */
      /* (otherwise the system call overhead will kill us)            */
      if (pnum % 1000 == 0)
	{
	  sigset_t mask;

	  sigemptyset (&mask);
	  sigaddset (&mask, SIGINT);

	  sigprocmask (SIG_UNBLOCK, &mask, NULL);
	  /* signal can happen EXACTLY HERE, when data structures are consistant */
	  sigprocmask (SIG_BLOCK, &mask, NULL);
	}
    }				/* end while == fine file */

  /* set ^C back to the default */
  /* (so we can kill the output if needed) */
  {
    sigset_t mask;

    sigemptyset (&mask);
    sigaddset (&mask, SIGINT);

    sigprocmask (SIG_UNBLOCK, &mask, NULL);
    signal (SIGINT, SIG_DFL);
  }

/* wait for the stat_dump thread to be idle */
  if (threaded)
    pthread_mutex_lock (&stat_dump_cond_mutex);
  else
    {
      sprintf (curr_data_dir, "%s/LAST", basename);
      if (debug > 1)
	printf ("DEB: writing stats for uncomplete traces... ");
      trace_done ();
      if (do_udp)
	udptrace_done ();

      if (debug > 1)
	printf ("DEB: writing addresses... ");

      swap_adx ();
      print_adx ();

      swap_histo ();		/*swappo perche lui stampa sempre i frozen */
      print_all_histo ();
    }

  /* close the input file */
  CompCloseFile (filename);
  dump_internal_stats (0);
}

void
QuitSig (int signum)
{
  printf ("%c\n\n", 7);		/* BELL */
  printf ("Terminating processing early on signal %d\n", signum);
  printf ("Partial result after processing %lu packets:\n\n\n", pnum);

  if (threaded)
    pthread_mutex_lock (&stat_dump_cond_mutex);

  sprintf (curr_data_dir, "%s/LAST", basename);

  if (debug > 1)
    printf ("DEB: writing addresses... ");
  swap_adx ();
  print_adx ();

  if (debug > 1)
    printf ("DEB: writing stats for uncomplete traces... ");
  trace_done ();
  if (do_udp)
    udptrace_done ();

  if (debug > 1)
    printf ("DEB: writing stats for complete traces... ");
  swap_histo ();
  print_all_histo ();

  dump_internal_stats (1);
  if (threaded)
    pthread_mutex_unlock (&stat_dump_cond_mutex);
  exit (EXIT_FAILURE);
}


#ifdef MEMDEBUG
void memory_debug ();
#endif
void
Usr1Sig (int signum)
{
  printf ("%c\n\n", 7);		/* BELL */
  printf ("Got a signal USR1\n");
  dump_internal_stats (1);
#ifdef MEMDEBUG
  memory_debug ();
#endif
}

void
dump_internal_stats (int err)
{
  double etime;
  gettimeofday (&wallclock_temp, NULL);
  etime = elapsed (wallclock_start, wallclock_temp);
  if (err)
    {
      fprintf (stderr, "\n\nDumping internal status variables:\n");
      fprintf (stderr, "total packet analized : %ld\n", pnum);
      fprintf (stderr, "total flows analized : %lu\n", fcount);
      fprintf (stderr, "total TCP flows analized : %lu\n", f_TCP_count);
      fprintf (stderr, "total UDP flows analized : %lu\n", f_UDP_count);
      fprintf (stderr, "total RTP flows analized : %lu\n", f_RTP_count);
      fprintf (stderr, "total RTCP flows analized : %lu\n", f_RTCP_count);
      fprintf (stderr,
	       "total iteration spent in the hash search routine : %d\n",
	       search_count);
      fprintf (stderr, "total analyzed TCP packet: %ld \n", tcp_packet_count);
      fprintf (stderr, "total analyzed UDP packet: %ld \n", udp_trace_count);

      fprintf (stderr, "total trash TCP packet: %ld \n", not_id_p);
      if (tcp_packet_count != 0)
	fprintf (stderr, "average TCP search length: %f\n",
		 (double) search_count / (double) tcp_packet_count);
      fprintf (stderr, "Current opened flows: TCP = %ld UDP = %ld\n",
	       tot_conn_TCP, tot_conn_UDP);
      fprintf (stderr, "Current flow vector index: %d (%d)\n", num_tcp_pairs,
	       MAX_TCP_PAIRS);
      fprintf (stderr, "Total adx used in hash: %ld \n", tot_adx_hash_count);
      fprintf (stderr, "Total adx used in list: %ld \n", tot_adx_list_count);
      fprintf (stderr, "Total adx hash search: %ld\n", adx_search_hash_count);
      fprintf (stderr, "Total adx list search: %ld\n", adx_search_list_count);
      printf ("elapsed wallclock time: %s\n", elapsed2str (etime));
      printf ("%d pkts/sec analyzed\n",
	      (int) ((double) pnum / (etime / 1000000)));
      printf ("%d flows/sec analyzed\n",
	      (int) ((double) fcount / (etime / 1000000)));
    }
  else
    {

      printf ("Dumping internal status variables:\n");
      printf ("total packet analized : %ld\n", pnum);
      printf ("total flows analized : %lu\n", fcount);
      printf ("total TCP flows analized : %lu\n", f_TCP_count);
      printf ("total UDP flows analized : %lu\n", f_UDP_count);
      printf ("total RTP flows analized : %lu\n", f_RTP_count);
      printf ("total RTCP flows analized : %lu\n", f_RTCP_count);
      printf ("total iteration spent in the hash search routine : %d\n",
	      search_count);
      printf ("total analyzed TCP packet: %ld \n", tcp_packet_count);
      printf ("total analyzed UDP packet: %ld \n", udp_trace_count);

      printf ("total trash TCP packet: %ld \n", not_id_p);
      if (tcp_packet_count != 0)
	printf ("average TCP search length: %ld\n",
		search_count / tcp_packet_count);
      printf ("Current opened flows: TCP = %ld UDP = %ld\n",
	       tot_conn_TCP, tot_conn_UDP);
      printf ("Current flow vector index: %d (%d)\n", num_tcp_pairs,
	      MAX_TCP_PAIRS);
      printf ("Total adx used in hash: %ld \n", tot_adx_hash_count);
      printf ("Total adx used in list: %ld \n", tot_adx_list_count);
      printf ("Total adx hash search: %ld\n", adx_search_hash_count);
      printf ("Total adx list search: %ld\n", adx_search_list_count);
      printf ("elapsed wallclock time: %s\n", elapsed2str (etime));
      printf ("%d pkts/sec analyzed\n",
	      (int) ((double) pnum / (etime / 1000000)));
      printf ("%d flows/sec analyzed\n",
	      (int) ((double) fcount / (etime / 1000000)));

#ifdef GROK_LIVE_TCPDUMP
      if (live_flag == TRUE && livecap_type == ETH)
	/* print out libpcap live capture stats */
	cleanup ();
#endif
    }

}

void *
MallocZ (int nbytes)
{
  char *ptr;

  ptr = malloc (nbytes);
  if (ptr == NULL)
    {
      perror ("Malloc failed, fatal\n");
      fprintf (stderr, "\
when memory allocation fails, it's either because:\n\
1) You're out of swap space, talk to your local sysadmin about making more\n\
   (look for system commands 'swap' or 'swapon' for quick fixes)\n\
2) The amount of memory that your OS gives each process is too little\n\
   That's a system configuration issue that you'll need to discuss\n\
   with the system administrator\n\
");
      exit (EXIT_FAILURE);
    }

  memset (ptr, 0, nbytes);	/* BZERO */

  return (ptr);
}

void *
ReallocZ (void *oldptr, int obytes, int nbytes)
{
  char *ptr;

  ptr = realloc (oldptr, nbytes);
  if (ptr == NULL)
    {
      fprintf (stderr,
	       "Realloc failed (%d bytes --> %d bytes), fatal\n",
	       obytes, nbytes);
      perror ("realloc");
      exit (EXIT_FAILURE);
    }
  if (obytes < nbytes)
    {
      memset ((char *) ptr + obytes, 0, nbytes - obytes);	/* BZERO */
    }

  return (ptr);
}


/* convert a buffer to an argc,argv[] pair */
void
StringToArgv (char *buf, int *pargc, char ***pargv)
{
  char **argv;
  int nargs = 0;

  /* discard the original string, use a copy */
  buf = strdup (buf);

  /* (very pessimistically) make the argv array */
  argv = malloc (sizeof (char *) * ((strlen (buf) / 2) + 1));

  /* skip leading blanks */
  while ((*buf != '\00') && (isspace ((int) *buf)))
    {
      if (debug > 10)
	printf ("skipping isspace('%c')\n", *buf);
      ++buf;
    }

  /* break into args */
  for (nargs = 1; *buf != '\00'; ++nargs)
    {
      char *stringend;
      argv[nargs] = buf;

      /* search for separator */
      while ((*buf != '\00') && (!isspace ((int) *buf)))
	{
	  if (debug > 10)
	    printf ("'%c' (%d) is NOT a space\n", *buf, (int) *buf);
	  ++buf;
	}
      stringend = buf;

      /* skip spaces */
      while ((*buf != '\00') && (isspace ((int) *buf)))
	{
	  if (debug > 10)
	    printf ("'%c' (%d) IS a space\n", *buf, (int) *buf);
	  ++buf;
	}

      *stringend = '\00';	/* terminate the previous string */

      if (debug)
	printf ("  argv[%d] = '%s'\n", nargs, argv[nargs]);
    }

  *pargc = nargs;
  *pargv = argv;
}


static void
CheckArguments (int *pargc, char *argv[])
{

  /* remember the name of the program for errors... */
  progname = argv[0];

  ParseArgs ("command line", pargc, argv);

  /* make sure we found the files */
  if (filenames == NULL && live_flag == FALSE
      && dump_all_histo_definition == FALSE)
    {
      BadArg (NULL, "must specify at least one file name\n");
    }
#ifdef HAVE_RRDTOOL
  /*-----------------------------------------------------------*/
  /* RRDtools                                                */
  /* make sure we found the files */
  if ((rrdset_path && !rrdset_conf) || (!rrdset_path && rrdset_conf))
    BadArg (NULL,
	    "You MUST specify both the configuration file (-R) AND the database path (-r))\n");
#endif
}


static void
ParseArgs (char *argsource, int *pargc, char *argv[])
{
  int i;
#ifdef GROK_ERF_LIVE
  int num_dev;
  int dim;
  char *ptr_help;
#endif

  /* parse the args */
  for (i = 1; i < *pargc; ++i)
    {
      if (argv[i] == NULL)
	continue;

#ifdef GROK_ERF_LIVE
      /* enable packet capture from DAG cards */
      if (strcmp (argv[i], "-dag") == 0)
	{
	  dim = DAG_NAME_BUFSIZE;
	  dag_dev_list = (char *) MallocZ (dim);
	  num_dev = 0;
	  live_flag = TRUE;
	  livecap_type = DAG;
	  while (argv[++i] != NULL && *argv[i] != '-')
	    {
	      if (num_dev == 4)
		{
		  fprintf (stderr,
			   "Error: are only supported at most four DAG card\n");
		  exit (1);
		}
	      else if (strlen (argv[i]) + strlen (dag_dev_list) > dim)
		{
		  dim += DAG_NAME_BUFSIZE;
		  ptr_help = (char *) MallocZ (dim);
		  strncpy (ptr_help, dag_dev_list, dim);
		  free (dag_dev_list);
		  dag_dev_list = ptr_help;
		}
	      strcat (dag_dev_list, argv[i]);
	      strcat (dag_dev_list, " ");
	      num_dev++;
	    }
	  if (num_dev == 0)
	    {
	      strncpy (dag_dev_list, "/dev/dag0 ", dim);	/* default DAG card */
	      num_dev++;
	    }
	  if (argv[i] == NULL)
	    continue;
	}
#endif /* GROK_ERF_LIVE */


      if (*argv[i] == '-')
	{
	  if (argv[i][1] == '\00')	/* just a '-' */
	    Usage ();

	  while (*(++argv[i]))
	    switch (*argv[i])
	      {
	      case 'N':
		if (*(argv[i] + 1))
		  {
		    /* -Nfile */
		    internal_net_file = strdup (argv[i] + 1);
		    *(argv[i] + 1) = '\00';
		    if (!LoadInternalNets (internal_net_file))
		      {
			fprintf (stderr,
				 "Error while loading configuration\n");
			fprintf (stderr, "Could not open %s\n",
				 internal_net_file);
			exit (1);
		      }
		  }
		else
		  {
		    /* maybe -N file */
		    BadArg (argsource, "-Nfile requires a file name\n");
		  }
		break;
	      case 'p':
		threaded = TRUE;
		break;
	      case 'd':
		++debug;
		break;
	      case 'h':
		Usage ();
		break;
	      case 'H':
		dump_all_histo_definition = TRUE;
		break;
#ifdef GROK_LIVE_TCPDUMP
	      case 'l':
		live_flag = TRUE;
		livecap_type = ETH;
		break;
#endif /* GROK_LIVE_TCPDUMP */
	      case 't':
		printticks = TRUE;
		break;
	      case 'u':
		do_udp = FALSE;
		break;
	      case 's':
		if (*(argv[i] + 1))
		  {
		    /* -sdir */
		    basenamedir = strdup (argv[i] + 1);
		    *(argv[i] + 1) = '\00';
		    basedirspecified = TRUE;
		    if (debug > 1)
		      printf ("basenamedir set to %s\n", basenamedir);
		  }
		else
		  {
		    /* maybe -s file */
		    BadArg (argsource, "-sdir requires a file name\n");
		  }
		break;
#ifdef GROK_LIVE_TCPDUMP
	      case 'i':	/* choose the live capture interface card */
		if (*(argv[i] + 1))
		  {
		    /* -ieth0 */
		    dev = strdup (argv[i] + 1);
		    *(argv[i] + 1) = '\00';
		    if (debug > 1)
		      printf ("Capturing device set to %s\n", dev);
		  }
		else
		  {
		    /* maybe -i dev */
		    BadArg (argsource, "-idev requires a device ID\n");
		  }
		break;
	      case 'f':	/*per mettere i filtri di cattura */
		if (*(argv[i] + 1))
		  {
		    /* -ffilter_file */
		    filter_filename = strdup (argv[i] + 1);
		    *(argv[i] + 1) = '\00';
		    filter_specified = TRUE;
		  }
		else
		  {
		    /* maybe -f filter_file */
		    BadArg (argsource, "-ffile requires a file\n");
		  }
		break;
#endif /* GROK_LIVE_TCPDUMP */
#ifdef HAVE_RRDTOOL
		/*-----------------------------------------------------------*/
		/* RRDtools                                                  */
	      case 'r':
		if (*(argv[i] + 1))
		  {		/* -Rfile */
		    char *rrdpath = strdup (argv[i] + 1);
		    struct stat fbuf;
		    *(argv[i] + 1) = '\00';
		    if ((stat (rrdpath, &fbuf) == 0)
			&& S_ISDIR (fbuf.st_mode))
		      {
			if (debug)
			  fprintf (stderr,
				   "RRDTool database path <%s> exists\n",
				   rrdpath);
		      }
		    else
		      {
			mkdir (rrdpath, 0775);
			if (debug)
			  fprintf (stderr,
				   "RRDTool database path <%s> created\n",
				   rrdpath);
		      }
		    rrdtool_set_path (rrdpath);
		    rrdset_path = 1;
		  }
		else
		  {
		    BadArg (argsource, "-rpath requires a directory name (and no space)\n");	/* maybe -N file */
		  }
		break;

	      case 'R':
		if (*(argv[i] + 1))
		  {		/* -Rfile */
		    char *rrdconf = strdup (argv[i] + 1);
		    struct stat fbuf;
		    *(argv[i] + 1) = '\00';
		    if (stat (rrdconf, &fbuf) == 0)
		      {
			if (debug)
			  fprintf (stderr,
				   "RRDTool configuration file <%s> found (delayed parsing)\n",
				   rrdconf);
			rrdtool_set_conf (rrdconf);
		      }
		    else
		      {
			fprintf (stderr, "Could not open %s\n", rrdconf);
			exit (1);
		      }
		    rrdset_conf = 1;
		  }
		else
		  {
		    BadArg (argsource, "-Rconf requires a file name (and no space)\n");	/* maybe -N file */
		  }
		break;
	      case 'S':
		pure_rrd_engine = TRUE;
		break;
		/*-----------------------------------------------------------*/
#endif
	      case 'v':
		Version ();
		exit (EXIT_SUCCESS);
		break;
	      case 'w':
		warn_printtrunc = TRUE;
		warn_printbadmbz = TRUE;
		warn_printbadcsum = TRUE;
		warn_printbad_syn_fin_seq = TRUE;
		warn_ooo = TRUE;
		break;
	      case '2':
		two_files = 2;
		break;
	      default:
		BadArg (argsource, "option '%c' not understood\n", *argv[i]);
	      }
	}
      else
	{
	  filenames = &argv[i];
	  *pargc -= i;
	  return;
	}
    }

  return;
}


int
LoadInternalNets (char *file)
{
  int i;
  FILE *fp;
  char c[50];

  fp = fopen (file, "r");
  if (fp == NULL)
    {
      return 0;
    }

  i = 0;

  while (fgets (c, 20, fp) != NULL)
    {
      inet_aton (c, &(internal_net_list[i]));
      if (fgets (c, 20, fp) != NULL)
	{
	  inet_aton (c, &(internal_net_mask2[i]));
	  internal_net_mask[i] = inet_addr (c);
	}
      else
	{
	  printf ("Cannot parse mask for net %d\n", (i + 1));
	  return 0;
	}
      if (debug)
	{
	  printf ("Adding: %s as internal net ",
		  inet_ntoa (internal_net_list[i]));
	  printf ("with mask %s (%d)\n", inet_ntoa (internal_net_mask2[i]),
		  internal_net_mask[i]);
	}
      i++;
    }
  tot_internal_nets = i;
  return 1;
}



/* the memcpy() function that gcc likes to stuff into the program has alignment
   problems, so here's MY version.  It's only used for small stuff, so the
   copy should be "cheap", but we can't be too fancy due to alignment boo boos */
void *
MemCpy (void *vp1, void *vp2, size_t n)
{
  char *p1 = vp1;
  char *p2 = vp2;

  while (n-- > 0)
    *p1++ = *p2++;

  return (vp1);
}


/* 
 * Check if the IP adx is included in the internal nets
 */

Bool
internal_ip (struct in_addr adx)
{
  int i;

  //printf("Checking %s \n",inet_ntoa(adx));
  for (i = 0; i < tot_internal_nets; i++)
    {
      //printf(" Against: %s \n",inet_ntoa(internal_net_list[i]));
      if ((adx.s_addr & internal_net_mask[i]) == internal_net_list[i].s_addr)
	{
	  //printf("Internal: %s\n",inet_ntoa(adx));
	  return 1;
	}
    }
  //printf("External: %s\n",inet_ntoa(adx));
  return 0;
}

/* 
 * Check if the IP adx is included in the internal nets
 * use the string dotted notation. Never used.
 */

Bool
internal_ip_string (char *adx)
{
  int i;

  for (i = 0; i < tot_internal_nets; i++)
    {
      if ((inet_addr (adx) & internal_net_mask[i]) ==
	  internal_net_list[i].s_addr)
	{
	  return 1;
	}
    }
  return 0;
}
